@@ -56,8 +56,10 @@ struct netreg_data {
 	int signal_index; /* If strength is reported via CIND */
 	int signal_min; /* min strength reported via CIND */
 	int signal_max; /* max strength reported via CIND */
+	int signal_invalid; /* invalid strength reported via CIND */
 	int tech;
 	struct ofono_network_time time;
+	guint nitz_timeout;
 	unsigned int vendor;
 };
 
@@ -665,7 +667,11 @@ static void ciev_notify(GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next_number(&iter, &strength))
 		return;
 
-	strength = (strength * 100) / (nd->signal_max - nd->signal_min);
+	if (strength == nd->signal_invalid)
+		strength = -1;
+	else
+		strength = (strength * 100) / (nd->signal_max - nd->signal_min);
+
 	ofono_netreg_strength_notify(netreg, strength);
 }
 
@@ -691,6 +697,18 @@ static void ctzv_notify(GAtResult *result, gpointer user_data)
 	ofono_netreg_time_notify(netreg, &nd->time);
 }
 
+static gboolean notify_time(gpointer user_data)
+{
+	struct ofono_netreg *netreg = user_data;
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+
+	nd->nitz_timeout = 0;
+
+	ofono_netreg_time_notify(netreg, &nd->time);
+
+	return FALSE;
+}
+
 static void ifx_ctzv_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_netreg *netreg = user_data;
@@ -722,6 +740,11 @@ static void ifx_ctzv_notify(GAtResult *result, gpointer user_data)
 	nd->time.mday = mday;
 	nd->time.mon = mon;
 	nd->time.year = 2000 + year;
+
+	if (nd->nitz_timeout > 0)
+		g_source_remove(nd->nitz_timeout);
+
+	nd->nitz_timeout = g_timeout_add_seconds(1, notify_time, user_data);
 }
 
 static void ifx_ctzdst_notify(GAtResult *result, gpointer user_data)
@@ -743,6 +766,11 @@ static void ifx_ctzdst_notify(GAtResult *result, gpointer user_data)
 
 	nd->time.dst = dst;
 
+	if (nd->nitz_timeout > 0) {
+		g_source_remove(nd->nitz_timeout);
+		nd->nitz_timeout = 0;
+	}
+
 	ofono_netreg_time_notify(netreg, &nd->time);
 }
 
@@ -775,7 +803,10 @@ static void cind_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	g_at_result_iter_next_number(&iter, &strength);
 
-	strength = (strength * 100) / (nd->signal_max - nd->signal_min);
+	if (strength == nd->signal_invalid)
+		strength = -1;
+	else
+		strength = (strength * 100) / (nd->signal_max - nd->signal_min);
 
 	cb(&error, strength, cbd->data);
 }
@@ -798,6 +829,50 @@ static void huawei_rssi_notify(GAtResult *result, gpointer user_data)
 				at_util_convert_signal_strength(strength));
 }
 
+static void huawei_nwtime_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_netreg *netreg = user_data;
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	int year, mon, mday, hour, min, sec;
+	char tz[4];
+	const char *date, *time, *dst;
+	GAtResultIter iter;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^NWTIME:"))
+		return;
+
+	if (!g_at_result_iter_next_unquoted_string(&iter, &date))
+		return;
+
+	if (!g_at_result_iter_next_unquoted_string(&iter, &time))
+		return;
+
+	if (!g_at_result_iter_next_unquoted_string(&iter, &dst))
+		return;
+
+	DBG("date %s time %s dst %s", date, time, dst);
+
+	if (sscanf(date, "%u/%u/%u", &year, &mon, &mday) != 3)
+		return;
+
+	if (sscanf(time, "%u:%u:%u%s", &hour, &min, &sec, tz) != 4)
+		return;
+
+	nd->time.utcoff = atoi(tz) * 15 * 60;
+	nd->time.dst = atoi(dst);
+
+	nd->time.sec = sec;
+	nd->time.min = min;
+	nd->time.hour = hour;
+	nd->time.mday = mday;
+	nd->time.mon = mon;
+	nd->time.year = 2000 + year;
+
+	ofono_netreg_time_notify(netreg, &nd->time);
+}
+
 static void csq_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
@@ -1110,7 +1185,9 @@ static void cind_support_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	GAtResultIter iter;
 	const char *str;
 	int index;
-	int min, max;
+	int min = 0;
+	int max = 0;
+	int tmp_min, tmp_max, invalid;
 
 	if (!ok)
 		goto error;
@@ -1121,15 +1198,30 @@ static void cind_support_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	index = 1;
 
+	/*
+	 * Telit encapsulates the CIND=? tokens with braces
+	 * so we need to skip them
+	 */
+	if (nd->vendor == OFONO_VENDOR_TELIT)
+		g_at_result_iter_open_list(&iter);
+
 	while (g_at_result_iter_open_list(&iter)) {
+		/* Reset invalid default value for every token */
+		invalid = 99;
+
 		if (!g_at_result_iter_next_string(&iter, &str))
 			goto error;
 
 		if (!g_at_result_iter_open_list(&iter))
 			goto error;
 
-		while (g_at_result_iter_next_range(&iter, &min, &max))
-			;
+		while (g_at_result_iter_next_range(&iter, &tmp_min, &tmp_max)) {
+			if (tmp_min != tmp_max) {
+				min = tmp_min;
+				max = tmp_max;
+			} else
+				invalid = tmp_min;
+		}
 
 		if (!g_at_result_iter_close_list(&iter))
 			goto error;
@@ -1141,11 +1233,15 @@ static void cind_support_cb(gboolean ok, GAtResult *result, gpointer user_data)
 			nd->signal_index = index;
 			nd->signal_min = min;
 			nd->signal_max = max;
+			nd->signal_invalid = invalid;
 		}
 
 		index += 1;
 	}
 
+	if (nd->vendor == OFONO_VENDOR_TELIT)
+		g_at_result_iter_close_list(&iter);
+
 	if (nd->signal_index == 0)
 		goto error;
 
@@ -1251,8 +1347,13 @@ static void at_creg_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 					FALSE, netreg, NULL);
 		break;
 	case OFONO_VENDOR_HUAWEI:
+		/* Register for RSSI reports */
 		g_at_chat_register(nd->chat, "^RSSI:", huawei_rssi_notify,
-					FALSE, netreg, NULL);
+						FALSE, netreg, NULL);
+
+		/* Register for network time reports */
+		g_at_chat_register(nd->chat, "^NWTIME:", huawei_nwtime_notify,
+						FALSE, netreg, NULL);
 		break;
 	case OFONO_VENDOR_IFX:
 		/* Register for specific signal strength reports */
@@ -1367,6 +1468,9 @@ static void at_netreg_remove(struct ofono_netreg *netreg)
 {
 	struct netreg_data *nd = ofono_netreg_get_data(netreg);
 
+	if (nd->nitz_timeout)
+		g_source_remove(nd->nitz_timeout);
+
 	ofono_netreg_set_data(netreg, NULL);
 
 	g_at_chat_unref(nd->chat);
