@@ -29,17 +29,20 @@
 #include <stdio.h>
 
 #include <glib.h>
+#include <gatchat.h>
+#include <gatresult.h>
 
 #include <ofono/log.h>
 #include <ofono/modem.h>
 #include <ofono/voicecall.h>
-#include <common.h>
-#include "gatchat.h"
-#include "gatresult.h"
+
+#include "common.h"
 
 #include "hfpmodem.h"
+#include "slc.h"
 
 #define POLL_CLCC_INTERVAL 2000
+#define POLL_CLCC_DELAY 50
 #define CLIP_TIMEOUT 500
 
 static const char *none_prefix[] = { NULL };
@@ -96,10 +99,12 @@ static struct ofono_call *create_call(struct ofono_voicecall *vc, int type,
 	struct ofono_call *call;
 
 	/* Generate a call structure for the waiting call */
-	call = g_try_new0(struct ofono_call, 1);
+	call = g_try_new(struct ofono_call, 1);
 	if (call == NULL)
 		return NULL;
 
+	ofono_call_init(call);
+
 	call->id = ofono_voicecall_get_next_callid(vc);
 	call->type = type;
 	call->direction = direction;
@@ -227,7 +232,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		if (oc && (nc == NULL || (nc->id > oc->id))) {
 			enum ofono_disconnect_reason reason;
 
-			if (vd->local_release & (0x1 << oc->id))
+			if (vd->local_release & (1 << oc->id))
 				reason = OFONO_DISCONNECT_REASON_LOCAL_HANGUP;
 			else
 				reason = OFONO_DISCONNECT_REASON_REMOTE_HANGUP;
@@ -335,7 +340,7 @@ static void atd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	for (l = vd->calls; l; l = l->next) {
 		call = l->data;
 
-		if (call->status != 0)
+		if (call->status != CALL_STATUS_ACTIVE)
 			continue;
 
 		call->status = CALL_STATUS_HELD;
@@ -355,16 +360,13 @@ out:
 
 static void hfp_dial(struct ofono_voicecall *vc,
 			const struct ofono_phone_number *ph,
-			enum ofono_clir_option clir, enum ofono_cug_option cug,
-			ofono_voicecall_cb_t cb, void *data)
+			enum ofono_clir_option clir, ofono_voicecall_cb_t cb,
+			void *data)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[256];
 
-	if (cbd == NULL)
-		goto error;
-
 	cbd->user = vc;
 	if (ph->type == 145)
 		snprintf(buf, sizeof(buf), "ATD+%s", ph->number);
@@ -377,7 +379,6 @@ static void hfp_dial(struct ofono_voicecall *vc,
 				atd_cb, cbd, g_free) > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
@@ -438,7 +439,7 @@ static void hfp_release_all_held(struct ofono_voicecall *vc,
 				ofono_voicecall_cb_t cb, void *data)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
-	unsigned int held_status = 0x1 << 1;
+	unsigned int held_status = 1 << CALL_STATUS_HELD;
 
 	if (vd->ag_mpty_features & AG_CHLD_0) {
 		hfp_template("AT+CHLD=0", vc, generic_cb, held_status,
@@ -453,7 +454,8 @@ static void hfp_set_udub(struct ofono_voicecall *vc,
 			ofono_voicecall_cb_t cb, void *data)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
-	unsigned int incoming_or_waiting = (0x1 << 4) | (0x1 << 5);
+	unsigned int incoming_or_waiting =
+		(1 << CALL_STATUS_INCOMING) | (1 << CALL_STATUS_WAITING);
 
 	if (vd->ag_mpty_features & AG_CHLD_0) {
 		hfp_template("AT+CHLD=0", vc, generic_cb, incoming_or_waiting,
@@ -649,7 +651,8 @@ static void ccwa_notify(GAtResult *result, gpointer user_data)
 
 	DBG("ccwa_notify: %s %d %d", num, num_type, validity);
 
-	call = create_call(vc, 0, 1, 5, num, num_type, validity);
+	call = create_call(vc, 0, 1, CALL_STATUS_WAITING, num, num_type,
+				validity);
 
 	if (call == NULL) {
 		ofono_error("malloc call struct failed.  "
@@ -941,7 +944,7 @@ static void ciev_callsetup_notify(struct ofono_voicecall *vc,
 		if (waiting == NULL && dialing == NULL)
 			goto out;
 
-		 /*
+		/*
 		 * If call=1, in the waiting case we have to poll, since we
 		 * have no idea whether a waiting call gave up or we accepted
 		 * using release+accept or hold+accept
@@ -1022,6 +1025,11 @@ static void ciev_callheld_notify(struct ofono_voicecall *vc,
 		break;
 
 	case 1:
+		if (vd->clcc_source) {
+			g_source_remove(vd->clcc_source);
+			vd->clcc_source = 0;
+		}
+
 		/* We have to poll here, we have no idea whether the call was
 		 * accepted by CHLD=1 or swapped by CHLD=2 or one call was
 		 * chosed for private chat by CHLD=2x
@@ -1041,7 +1049,15 @@ static void ciev_callheld_notify(struct ofono_voicecall *vc,
 				ofono_voicecall_notify(vc, call);
 			}
 		} else if (callheld == 1) {
-			release_with_status(vc, CALL_STATUS_ACTIVE);
+			if (vd->clcc_source)
+				g_source_remove(vd->clcc_source);
+
+			/* We have to schedule a poll here, we have no idea
+			 * whether active call was dropped by remote or if this
+			 * is an intermediate state during call swap
+			 */
+			vd->clcc_source = g_timeout_add(POLL_CLCC_DELAY,
+							poll_clcc, vc);
 		}
 	}
 
@@ -1115,17 +1131,17 @@ static void hfp_voicecall_initialized(gboolean ok, GAtResult *result,
 static int hfp_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
 				gpointer user_data)
 {
-	struct hfp_data *data = user_data;
+	struct hfp_slc_info *info = user_data;
 	struct voicecall_data *vd;
 
 	vd = g_new0(struct voicecall_data, 1);
 
-	vd->chat = data->chat;
-	vd->ag_features = data->ag_features;
-	vd->ag_mpty_features = data->ag_mpty_features;
+	vd->chat = g_at_chat_clone(info->chat);
+	vd->ag_features = info->ag_features;
+	vd->ag_mpty_features = info->ag_mpty_features;
 
-	memcpy(vd->cind_pos, data->cind_pos, HFP_INDICATOR_LAST);
-	memcpy(vd->cind_val, data->cind_val, HFP_INDICATOR_LAST);
+	memcpy(vd->cind_pos, info->cind_pos, HFP_INDICATOR_LAST);
+	memcpy(vd->cind_val, info->cind_val, HFP_INDICATOR_LAST);
 
 	ofono_voicecall_set_data(vc, vd);
 
