@@ -60,6 +60,9 @@ typedef void (*stk_agent_string_cb)(enum stk_agent_result result,
 typedef void (*stk_agent_tone_cb)(enum stk_agent_result result,
 						void *user_data);
 
+typedef void (*stk_agent_display_action_cb)(enum stk_agent_result result,
+						void *user_data);
+
 struct stk_agent *stk_agent_new(const char *path, const char *sender,
 					ofono_bool_t remove_on_terminate);
 
@@ -147,3 +150,16 @@ int stk_agent_confirm_launch_browser(struct stk_agent *agent, const char *text,
 					void *user_data,
 					ofono_destroy_func destroy,
 					int timeout);
+
+int stk_agent_display_action(struct stk_agent *agent, const char *text,
+					const struct stk_icon_id *icon,
+					stk_agent_display_action_cb cb,
+					void *user_data,
+					ofono_destroy_func destroy);
+
+int stk_agent_confirm_open_channel(struct stk_agent *agent, const char *text,
+					const struct stk_icon_id *icon,
+					stk_agent_confirmation_cb cb,
+					void *user_data,
+					ofono_destroy_func destroy,
+					int timeout);
